/*!
 * Copyright (c) 2022 by XGBoost Contributors
 */
#ifndef XGBOOST_TESTS_CPP_FILESYSTEM_H
#define XGBOOST_TESTS_CPP_FILESYSTEM_H

// A macro used inside `windows.h` to avoid conflicts with `winsock2.h`
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif  // WIN32_LEAN_AND_MEAN

#include "dmlc/filesystem.h"

#endif  // XGBOOST_TESTS_CPP_FILESYSTEM_H
