/*!
 * Copyright 2022 XGBoost contributors
 */
#include "in_memory_communicator.h"

namespace xgboost {
namespace collective {

InMemoryHandler InMemoryCommunicator::handler_{};

}  // namespace collective
}  // namespace xgboost
